module code.firrtl.Syntax where

-- This is the syntax for Idealized FIRRTL

open import Relation.Binary.PropositionalEquality hiding ([_])

open import Data.Bool renaming (Bool to Bit; false to 0#; true to 1#)
open import Data.Fin using (Fin; zero; suc)
open import Data.List hiding (lookup ; [_] ; and)
open import Data.Maybe hiding (map)
open import Data.Nat using (ℕ; _∸_; _<_ ; _≤_ ; _+_) renaming (zero to Z; suc to S)
open import Data.Product hiding (map ; swap ; curry ; uncurry)
open import Data.Vec hiding (foldl ; foldr ; _++_ ; map ; _[_]=_ ; length) renaming (lookup to vecref)

open import code.common.Idioms hiding (_∘_)
open import code.common.Contexts

open import code.firrtl.Types

--------------
-- New definitions
--------------

-- ≻ = slash succ
-- ⋉ = slash ltimes
-- ∝ = slash propto
-- ▪ = ?

ins : ModTy → Ty
ins (Mod i , s , o ⋊ μ)   = i

regs : ModTy → Ty
regs (Mod i , s , o ⋊ μ)  = s

outs : ModTy → Ty
outs (Mod i , s , o ⋊ μ)  = o

insts : ModTy → List ModTy
insts (Mod i , s , o ⋊ μ) = μ

prodTy : List Ty → Ty
prodTy []       = Unit
prodTy (d ∷ []) = d
prodTy (d ∷ ds) = Πty d ds

instsᵢ : ModTy → Ty
instsᵢ d = prodTy (Data.List.map ins (insts d)) 

lemmaPT : (t : Ty) → (ts : List Ty) → (prodTy (t ∷ ts)) ≡ (Πty t ts) 
lemmaPT t []       = refl
lemmaPT t (x ∷ ts) = refl

instsₒ : ModTy → Ty
instsₒ d = prodTy (Data.List.map outs (insts d))

visible : ModTy → Cxt Ty
visible Δ = instsₒ Δ ∷ regs Δ ∷ ins Δ ∷ []

data _⊢ₓ_//_ : Cxt GrTy → GrTy → ModTy → Set where

  reg    : ∀ {Γ Δ a}  
         → a ∈ᵍ regs Δ 
           -----------
         →  Γ ⊢ₓ a // Δ

  inp    : ∀ {Γ Δ a}  
         → a ∈ᵍ ins Δ 
           -----------
         → Γ ⊢ₓ a // Δ

  node    : ∀ {Γ Δ a}
          → a ∈ Γ     
            -----------
          → Γ ⊢ₓ a // Δ
           
  --
  -- de Bruijn-ed qualified variables
  --   N.b., for a module instance m, m's outputs are read-only, but 
  --         m's inputs are write-only. This is the opposite of locally
  --         declared ports.
  --
  qexp   : ∀ {Γ Δ a} 
         →  (μ : Ty)   
         → μ ∈ᵗ instsₒ Δ
         → a ∈ᵍ μ       
           -------------------
         → Γ ⊢ₓ a // Δ

  --
  -- Literals
  -- 

  uint :   ∀ {Γ Δ n}     →
           Vec Bit n     →
           ------------------
           Γ ⊢ₓ (UInt n) // Δ

  sint :   ∀ {Γ Δ n} →
           Vec Bit n →
           ------------------
           Γ ⊢ₓ (SInt n) // Δ

  xtail :  ∀ {Γ Δ n}                →
           (m : ℕ)                  →
           Γ ⊢ₓ (UInt (n + m)) // Δ →
           ------------------------
           Γ ⊢ₓ UInt m // Δ

  mux :    ∀ {Γ Δ τ} 
      →    Γ ⊢ₓ UInt 1 // Δ 
      →    Γ ⊢ₓ τ // Δ      
      →    Γ ⊢ₓ τ // Δ      
           -------------
      →    Γ ⊢ₓ τ // Δ

  and :    ∀ {Γ Δ n}        →
           Γ ⊢ₓ UInt n // Δ →
           Γ ⊢ₓ UInt n // Δ →
           -------------
           Γ ⊢ₓ UInt n // Δ

  add :    ∀ {Γ Δ n}        →
           Γ ⊢ₓ UInt n // Δ →
           Γ ⊢ₓ UInt n // Δ →
           -------------
           Γ ⊢ₓ UInt n // Δ


  -- bits   : ∀ {i2 o1 r2 Γ m hi lo}
  --        → i2 , o1 , r2 , Γ ⊢ᵉ UInt m
  --        → hi ≤ m   -- ungainly, but it works. Somehow, decidable 
  --        → lo ≤ hi  -- relations could simplify this.
  --          -------------
  --        → i2 , o1 , r2 , Γ ⊢ᵉ UInt ((hi ∸ lo) + 1)

--
-- These are "bodies"; i.e.,
--    modules ::= module ins regs outs insts nodes body
--    body    ::= 
--
data _⊢ᵇ_//_ : Cxt GrTy → Ty → ModTy → Set where

    unit : ∀ {Γ Δ}         →
             --------------
             Γ ⊢ᵇ Unit // Δ
             
    ⊗case : ∀ {Γ Δ t ts} →
              Γ ⊢ₓ t // Δ         →
              Γ ⊢ᵇ ts // Δ        →
              --------------------
              Γ ⊢ᵇ (↑ t) ⊗ ts // Δ

------------------------------------------------------
-- Node Rules
------------------------------------------------------

-- structural rules for checking lists of expressions
data _⊢ₓ̂_//_ : Cxt GrTy → List GrTy → ModTy → Set where

    none : ∀ {Γ Δ} →
           ------------
           Γ ⊢ₓ̂ [] // Δ

    cons : ∀ {Γ Δ t ts} →
           Γ ⊢ₓ t // Δ  →
           Γ ⊢ₓ̂ ts // Δ →
           ------------
           Γ ⊢ₓ̂ (t ∷ ts) // Δ


data _⊢ⁿ_//_ : Cxt GrTy → Ty → ModTy → Set where

    node : ∀ {Γ Δ t as}       →
           Γ ⊢ₓ̂ as // Δ        → -- let-bound nodes
           (as ++ Γ) ⊢ᵇ t // Δ → 
           -------------------
           (as ++ Γ) ⊢ⁿ t // Δ

------------------------------------------------------
-- Module Rules
------------------------------------------------------

infix  4  _⊢ᴹ_
infix  4  _⊢ᵐ_

--
-- Records a sequence of relevant module references as instance declarations.
--
data _⊢ᴹ_ : Cxt ModTy → List ModTy → Set where
  ◇   : ∀ {Μ} → Μ ⊢ᴹ []
  _◁_ : ∀ {Μ d ds} → d ∈ Μ → Μ ⊢ᴹ ds → Μ ⊢ᴹ (d ∷ ds)

data _⊢ᵐ_ : Cxt ModTy → ModTy → Set where
  mod : ∀ {Μ Δ}              →
        (Μ ⊢ᴹ insts Δ)       → -- rather than insts Δ ⊆ Μ
        ([] ⊢ⁿ outs Δ // Δ)  → -- out connected
        ([] ⊢ⁿ regs Δ // Δ)  → -- regs connected
        ([] ⊢ⁿ instsᵢ Δ // Δ) → -- internal inputs connected
        -------------------
        Μ ⊢ᵐ Δ 

-----------------------------------------------------------
-- Circuit Rules
--    A circuit, by definition, is a closed term, [] ⊢ᶜ Δ.
-----------------------------------------------------------

infix  4  _⊢ᶜ_

data _⊢ᶜ_ : Cxt ModTy → ModTy → Set where

   mod  : ∀ {Μ Δ}
        →  Μ ⊢ᵐ Δ  
          -------
        → Μ ⊢ᶜ Δ  

   letmod : ∀ {Μ Δ'}   
          → (Δ : ModTy)  
          → (Μ ⊢ᵐ Δ)    
          → (Δ ∷ Μ) ⊢ᶜ Δ' 
            -------
          → Μ ⊢ᶜ Δ'
